#include <bits/stdc++.h>
#include <sys/stat.h>
#include <unistd.h>
#include <iostream>
#include <fcntl.h>
using namespace std;

using lsn_t = int32_t;
using txn_id_t = int32_t;

static constexpr int INVALID_FRAME_ID = -1;                                   // invalid frame id
static constexpr int INVALID_PAGE_ID = -1;                                    // invalid page id
static constexpr int INVALID_TXN_ID = -1;                                     // invalid transaction id
static constexpr int INVALID_TIMESTAMP = -1;                                  // invalid transaction timestamp
static constexpr int INVALID_LSN = -1;                                        // invalid log sequence number
static constexpr int HEADER_PAGE_ID = 0;                                      // the header page id
static constexpr int PAGE_SIZE = 4096; 

static constexpr int OFFSET_LOG_TYPE = 0;
// the offset of lsn_ in log header
static constexpr int OFFSET_LSN = sizeof(int);
// the offset of log_tot_len_ in log header
static constexpr int OFFSET_LOG_TOT_LEN = OFFSET_LSN + sizeof(lsn_t);
// the offset of log_tid_ in log header
static constexpr int OFFSET_LOG_TID = OFFSET_LOG_TOT_LEN + sizeof(uint32_t);
// the offset of prev_lsn_ in log header
static constexpr int OFFSET_PREV_LSN = OFFSET_LOG_TID + sizeof(txn_id_t);
// offset of log data
static constexpr int OFFSET_LOG_DATA = OFFSET_PREV_LSN + sizeof(lsn_t);
// sizeof log_header
static constexpr int LOG_HEADER_SIZE = OFFSET_LOG_DATA;

static constexpr int LOG_BUFFER_SIZE = (1024 * PAGE_SIZE);

enum LogType: int {
    UPDATE = 0,
    INSERT,
    DELETE,
    begin,
    commit,
    ABORT,
    IX_INSERT,
    IX_DELETE
};
static std::string LogTypeStr[] = {
    "UPDATE",
    "INSERT",
    "DELETE",
    "BEGIN",
    "COMMIT",
    "ABORT"
};

struct Rid {
    int page_no;
    int slot_no;

    friend bool operator==(const Rid &x, const Rid &y) {
        return x.page_no == y.page_no && x.slot_no == y.slot_no;
    }

    friend bool operator!=(const Rid &x, const Rid &y) { return !(x == y); }

    void copy(const Rid& value) {
        page_no = value.page_no;
        slot_no = value.slot_no;
    }
};

struct RmRecord {
    char* data;  // 记录的数据
    int size;    // 记录的大小
    bool allocated_ = false;    // 是否已经为数据分配空间

    RmRecord() = default;

    RmRecord(const RmRecord& other) {
        size = other.size;
        data = new char[size];
        memcpy(data, other.data, size);
        allocated_ = true;
    };

    RmRecord &operator=(const RmRecord& other) {
        size = other.size;
        data = new char[size];
        memcpy(data, other.data, size);
        allocated_ = true;
        return *this;
    };

    RmRecord(int size_) {
        size = size_;
        data = new char[size_];
        allocated_ = true;
    }

    RmRecord(int size_, char* data_) {
        size = size_;
        data = new char[size_];
        memcpy(data, data_, size_);
        allocated_ = true;
    }

    void SetData(char* data_) {
        memcpy(data, data_, size);
    }

    void Deserialize(const char* data_) {
        size = *reinterpret_cast<const int*>(data_);
        if(allocated_) {
            delete[] data;
        }
        data = new char[size];
        memcpy(data, data_ + sizeof(int), size);
    }


    ~RmRecord() {
        if(allocated_) {
            delete[] data;
        }
        allocated_ = false;
        data = nullptr;
    }
};

class LogRecord {
public:
    LogType log_type_;         /* 日志对应操作的类型 */
    lsn_t lsn_;                /* 当前日志的lsn */
    uint32_t log_tot_len_;     /* 整个日志记录的长度 */
    txn_id_t log_tid_;         /* 创建当前日志的事务ID */
    lsn_t prev_lsn_;           /* 事务创建的前一条日志记录的lsn，用于undo */

    // 把日志记录序列化到dest中
    virtual void serialize (char* dest) const {
        memcpy(dest + OFFSET_LOG_TYPE, &log_type_, sizeof(LogType));
        memcpy(dest + OFFSET_LSN, &lsn_, sizeof(lsn_t));
        memcpy(dest + OFFSET_LOG_TOT_LEN, &log_tot_len_, sizeof(uint32_t));
        memcpy(dest + OFFSET_LOG_TID, &log_tid_, sizeof(txn_id_t));
        memcpy(dest + OFFSET_PREV_LSN, &prev_lsn_, sizeof(lsn_t));
    }
    // 从src中反序列化出一条日志记录
    virtual void deserialize(const char* src) {
        log_type_ = *reinterpret_cast<const LogType*>(src);
        lsn_ = *reinterpret_cast<const lsn_t*>(src + OFFSET_LSN);
        log_tot_len_ = *reinterpret_cast<const uint32_t*>(src + OFFSET_LOG_TOT_LEN);
        log_tid_ = *reinterpret_cast<const txn_id_t*>(src + OFFSET_LOG_TID);
        prev_lsn_ = *reinterpret_cast<const lsn_t*>(src + OFFSET_PREV_LSN);
    }
    // used for debug
    virtual void format_print() {
        std::cout << "log type in father_function: " << LogTypeStr[log_type_] << "\n";
        printf("Print Log Record:\n");
        printf("log_type_: %s\n", LogTypeStr[log_type_].c_str());
        printf("lsn: %d\n", lsn_);
        printf("log_tot_len: %d\n", log_tot_len_);
        printf("log_tid: %d\n", log_tid_);
        printf("prev_lsn: %d\n", prev_lsn_);
    }
};


class BeginLogRecord: public LogRecord {
public:
    BeginLogRecord() {
        log_type_ = LogType::begin;
        lsn_ = INVALID_LSN;
        log_tot_len_ = LOG_HEADER_SIZE;
        log_tid_ = INVALID_TXN_ID;
        prev_lsn_ = INVALID_LSN;
    }
    BeginLogRecord(txn_id_t txn_id) : BeginLogRecord() {
        log_tid_ = txn_id;
    }
    // 序列化Begin日志记录到dest中
    void serialize(char* dest) const override {
        LogRecord::serialize(dest);
    }
    // 从src中反序列化出一条Begin日志记录
    void deserialize(const char* src) override {
        LogRecord::deserialize(src);   
    }
    virtual void format_print() override {
        std::cout << "log type in son_function: " << LogTypeStr[log_type_] << "\n";
        LogRecord::format_print();
    }
};

/**
 * TODO: commit操作的日志记录
*/
class CommitLogRecord: public LogRecord {
public:
    CommitLogRecord() {
        log_type_ = LogType::commit;
        lsn_ = INVALID_LSN;
        log_tot_len_ = LOG_HEADER_SIZE;
        log_tid_ = INVALID_TXN_ID;
        prev_lsn_ = INVALID_LSN;
    }
    CommitLogRecord(txn_id_t txn_id) : CommitLogRecord() {
        log_tid_ = txn_id;
    }
    // 序列化commit日志记录到dest中
    void serialize(char* dest) const override {
        LogRecord::serialize(dest);
    }
    // 从src中反序列化出一条commit日志记录
    void deserialize(const char* src) override {
        LogRecord::deserialize(src);   
    }
    virtual void format_print() override {
        std::cout << "log type in son_function: " << LogTypeStr[log_type_] << "\n";
        LogRecord::format_print();
    }
};

/**
 * TODO: abort操作的日志记录
*/
class AbortLogRecord: public LogRecord {
public:
    AbortLogRecord() {
        log_type_ = LogType::ABORT;
        lsn_ = INVALID_LSN;
        log_tot_len_ = LOG_HEADER_SIZE;
        log_tid_ = INVALID_TXN_ID;
        prev_lsn_ = INVALID_LSN;
    }
    AbortLogRecord(txn_id_t txn_id) : AbortLogRecord() {
        log_tid_ = txn_id;
    }
    // 序列化abort日志记录到dest中
    void serialize(char* dest) const override {
        LogRecord::serialize(dest);
    }
    // 从src中反序列化出一条abort日志记录
    void deserialize(const char* src) override {
        LogRecord::deserialize(src);   
    }
    virtual void format_print() override {
        std::cout << "log type in son_function: " << LogTypeStr[log_type_] << "\n";
        LogRecord::format_print();
    }
};

class InsertLogRecord: public LogRecord {
public:
    InsertLogRecord() {
        log_type_ = LogType::INSERT;
        lsn_ = INVALID_LSN;
        log_tot_len_ = LOG_HEADER_SIZE;
        log_tid_ = INVALID_TXN_ID;
        prev_lsn_ = INVALID_LSN;
        table_name_ = nullptr;
    }
    InsertLogRecord(txn_id_t txn_id, RmRecord& insert_value, Rid& rid, std::string table_name) 
        : InsertLogRecord() {
        log_tid_ = txn_id;
        insert_value_ = insert_value;
        rid_ = rid;
        log_tot_len_ += sizeof(int);
        log_tot_len_ += insert_value_.size;
        log_tot_len_ += sizeof(Rid);
        table_name_size_ = table_name.length();
        table_name_ = new char[table_name_size_];
        memcpy(table_name_, table_name.c_str(), table_name_size_);
        log_tot_len_ += sizeof(size_t) + table_name_size_;
    }

    // 把insert日志记录序列化到dest中
    void serialize(char* dest) const override {
        LogRecord::serialize(dest);
        int offset = OFFSET_LOG_DATA;
        memcpy(dest + offset, &insert_value_.size, sizeof(int));
        offset += sizeof(int);
        memcpy(dest + offset, insert_value_.data, insert_value_.size);
        offset += insert_value_.size;
        memcpy(dest + offset, &rid_, sizeof(Rid));
        offset += sizeof(Rid);
        memcpy(dest + offset, &table_name_size_, sizeof(size_t));
        offset += sizeof(size_t);
        memcpy(dest + offset, table_name_, table_name_size_);
    }
    // 从src中反序列化出一条Insert日志记录
    void deserialize(const char* src) override {
        LogRecord::deserialize(src);  
        insert_value_.Deserialize(src + OFFSET_LOG_DATA);
        int offset = OFFSET_LOG_DATA + insert_value_.size + sizeof(int);
        rid_ = *reinterpret_cast<const Rid*>(src + offset);
        offset += sizeof(Rid);
        table_name_size_ = *reinterpret_cast<const size_t*>(src + offset);
        offset += sizeof(size_t);
        table_name_ = new char[table_name_size_];
        memcpy(table_name_, src + offset, table_name_size_);
    }
    void format_print() override {
        printf("insert record\n");
        LogRecord::format_print();
        printf("insert_value: %s\n", insert_value_.data);
        printf("insert rid: %d, %d\n", rid_.page_no, rid_.slot_no);
        printf("table name: %s\n", table_name_);
    }

    std::string get_tab_name() {
        std::string ret(table_name_, table_name_size_);
        return ret;
    }

    RmRecord insert_value_;     // 插入的记录
    Rid rid_;                   // 记录插入的位置
    char* table_name_;          // 插入记录的表名称
    size_t table_name_size_;    // 表名称的大小
};

/**
 * TODO: delete操作的日志记录
*/
class DeleteLogRecord: public LogRecord {
public:
    DeleteLogRecord() {
        log_type_ = LogType::DELETE;
        lsn_ = INVALID_LSN;
        log_tot_len_ = LOG_HEADER_SIZE;
        log_tid_ = INVALID_TXN_ID;
        prev_lsn_ = INVALID_LSN;
        table_name_ = nullptr;
    }
    DeleteLogRecord(txn_id_t txn_id, RmRecord& delete_value, Rid& rid, std::string table_name) 
        : DeleteLogRecord() {
        log_tid_ = txn_id;
        delete_value_ = delete_value;
        rid_ = rid;
        log_tot_len_ += sizeof(int);
        log_tot_len_ += delete_value_.size;
        log_tot_len_ += sizeof(Rid);
        table_name_size_ = table_name.length();
        table_name_ = new char[table_name_size_];
        memcpy(table_name_, table_name.c_str(), table_name_size_);
        log_tot_len_ += sizeof(size_t) + table_name_size_;
    }

    // 把delete日志记录序列化到dest中
    void serialize(char* dest) const override {
        LogRecord::serialize(dest);
        int offset = OFFSET_LOG_DATA;
        memcpy(dest + offset, &delete_value_.size, sizeof(int));
        offset += sizeof(int);
        memcpy(dest + offset, delete_value_.data, delete_value_.size);
        offset += delete_value_.size;
        memcpy(dest + offset, &rid_, sizeof(Rid));
        offset += sizeof(Rid);
        memcpy(dest + offset, &table_name_size_, sizeof(size_t));
        offset += sizeof(size_t);
        memcpy(dest + offset, table_name_, table_name_size_);
    }
    // 从src中反序列化出一条delete日志记录
    void deserialize(const char* src) override {
        LogRecord::deserialize(src);  
        delete_value_.Deserialize(src + OFFSET_LOG_DATA);
        int offset = OFFSET_LOG_DATA + delete_value_.size + sizeof(int);
        rid_ = *reinterpret_cast<const Rid*>(src + offset);
        offset += sizeof(Rid);
        table_name_size_ = *reinterpret_cast<const size_t*>(src + offset);
        offset += sizeof(size_t);
        table_name_ = new char[table_name_size_];
        memcpy(table_name_, src + offset, table_name_size_);
    }
    void format_print() override {
        printf("delete record\n");
        LogRecord::format_print();
        printf("delete_value: %s\n", delete_value_.data);
        printf("delete rid: %d, %d\n", rid_.page_no, rid_.slot_no);
        printf("table name: %s\n", table_name_);
    }

    std::string get_tab_name() {
        std::string ret(table_name_, table_name_size_);
        return ret;
    }

    RmRecord delete_value_;     // 插入的记录
    Rid rid_;                   // 记录插入的位置
    char* table_name_;          // 插入记录的表名称
    size_t table_name_size_;    // 表名称的大小
};

/**
 * TODO: update操作的日志记录
*/
class UpdateLogRecord: public LogRecord {
public:
    UpdateLogRecord() {
        log_type_ = LogType::UPDATE;
        lsn_ = INVALID_LSN;
        log_tot_len_ = LOG_HEADER_SIZE;
        log_tid_ = INVALID_TXN_ID;
        prev_lsn_ = INVALID_LSN;
        table_name_ = nullptr;
    }
    UpdateLogRecord(txn_id_t txn_id, RmRecord& update_value, RmRecord& old_value, Rid& rid, std::string table_name) 
        : UpdateLogRecord() {
        log_tid_ = txn_id;
        old_value_ = old_value;
        update_value_ = update_value;
        rid_ = rid;
        log_tot_len_ += sizeof(int);
        log_tot_len_ += update_value.size;
        log_tot_len_ += old_value.size;
        log_tot_len_ += sizeof(int);
        log_tot_len_ += sizeof(Rid);
        table_name_size_ = table_name.length();
        table_name_ = new char[table_name_size_];
        memcpy(table_name_, table_name.c_str(), table_name_size_);
        log_tot_len_ += sizeof(size_t) + table_name_size_;
    }

    // 把update日志记录序列化到dest中
    void serialize(char* dest) const override {
        //log头
        LogRecord::serialize(dest);
        int offset = OFFSET_LOG_DATA;

        //旧值
        memcpy(dest + offset, &old_value_.size, sizeof(int));
        offset += sizeof(int);
        memcpy(dest + offset, old_value_.data, old_value_.size);
        offset += old_value_.size;

        //新值
        memcpy(dest + offset, &update_value_.size, sizeof(int));
        offset += sizeof(int);
        memcpy(dest + offset, update_value_.data, update_value_.size);
        offset += update_value_.size;

        memcpy(dest + offset, &rid_, sizeof(Rid));
        offset += sizeof(Rid);
        memcpy(dest + offset, &table_name_size_, sizeof(size_t));
        offset += sizeof(size_t);
        memcpy(dest + offset, table_name_, table_name_size_);
    }
    // 从src中反序列化出一条update日志记录
    void deserialize(const char* src) override {
        //得到头
        LogRecord::deserialize(src);  

        //得到旧值
        old_value_.Deserialize(src + OFFSET_LOG_DATA);
        int offset = OFFSET_LOG_DATA + old_value_.size + sizeof(int);

        //得到新值
        update_value_.Deserialize(src + offset);
        offset += update_value_.size + sizeof(int);

        rid_ = *reinterpret_cast<const Rid*>(src + offset);
        offset += sizeof(Rid);
        table_name_size_ = *reinterpret_cast<const size_t*>(src + offset);
        offset += sizeof(size_t);
        table_name_ = new char[table_name_size_];
        memcpy(table_name_, src + offset, table_name_size_);
    }
    void format_print() override {
        printf("update record\n");
        LogRecord::format_print();
        printf("old_value: %s\n", old_value_.data);
        printf("update_value: %s\n", update_value_.data);
        printf("update rid: %d, %d\n", rid_.page_no, rid_.slot_no);
        printf("table name: %s\n", table_name_);
    }

    std::string get_tab_name() {
        std::string ret(table_name_, table_name_size_);
        return ret;
    }

    RmRecord old_value_;        // 旧的记录
    RmRecord update_value_;     // 更新的记录
    Rid rid_;                   // 记录的位置
    char* table_name_;          // 记录的表名称
    size_t table_name_size_;    // 表名称的大小
};


/* 日志缓冲区，只有一个buffer，因此需要阻塞地去把日志写入缓冲区中 */

class LogBuffer {
public:
    LogBuffer() { 
        offset_ = 0; 
        memset(buffer_, 0, sizeof(buffer_));
    }

    void resetBuffer(){
        offset_ = 0; 
    }

    bool is_full(int append_size) {
        if(offset_ + append_size > LOG_BUFFER_SIZE)
            return true;
        return false;
    }

    char buffer_[LOG_BUFFER_SIZE+1];
    int offset_;    // 写入log的offset
};

class IxInsertLogRecord: public LogRecord {
public:
    IxInsertLogRecord() {
        log_type_ = LogType::IX_INSERT;
        lsn_ = INVALID_LSN;
        log_tot_len_ = LOG_HEADER_SIZE;
        log_tid_ = INVALID_TXN_ID;
        prev_lsn_ = INVALID_LSN;
        index_file_name_ = nullptr;
    }
    IxInsertLogRecord(txn_id_t txn_id, const char *key, int key_size, const Rid &rid, std::string table_name) 
        : IxInsertLogRecord() {
        log_tid_ = txn_id;

        insert_value_.size = key_size;
        insert_value_.data = new char[key_size];
        memcpy(insert_value_.data, key, key_size);
        rid_ = rid;
        log_tot_len_ += sizeof(int);
        log_tot_len_ += sizeof(Rid);
        log_tot_len_ += insert_value_.size;
        index_file_name_size_ = table_name.length();
        index_file_name_ = new char[index_file_name_size_];
        memcpy(index_file_name_, table_name.c_str(), index_file_name_size_);
        log_tot_len_ += sizeof(size_t) + index_file_name_size_;
    }

    // 把insert日志记录序列化到dest中
    void serialize(char* dest) const override {
        LogRecord::serialize(dest);
        int offset = OFFSET_LOG_DATA;
        memcpy(dest + offset, &insert_value_.size, sizeof(int));
        offset += sizeof(int);
        memcpy(dest + offset, insert_value_.data, insert_value_.size);
        offset += insert_value_.size;
        memcpy(dest + offset, &rid_, sizeof(Rid));
        offset += sizeof(Rid);
        memcpy(dest + offset, &index_file_name_size_, sizeof(size_t));
        offset += sizeof(size_t);
        memcpy(dest + offset, index_file_name_, index_file_name_size_);
    }
    // 从src中反序列化出一条Insert日志记录
    void deserialize(const char* src) override {
        LogRecord::deserialize(src);  
        insert_value_.Deserialize(src + OFFSET_LOG_DATA);
        int offset = OFFSET_LOG_DATA + insert_value_.size + sizeof(int);
        rid_ = *reinterpret_cast<const Rid*>(src + offset);
        offset += sizeof(Rid);
        index_file_name_size_ = *reinterpret_cast<const size_t*>(src + offset);
        offset += sizeof(size_t);
        index_file_name_ = new char[index_file_name_size_];
        memcpy(index_file_name_, src + offset, index_file_name_size_);
    }
    void format_print() override {
        printf("insert key\n");
        LogRecord::format_print();
        printf("insert_value_: %s\n", insert_value_.data);
        printf("record rid: %d, %d\n", rid_.page_no, rid_.slot_no);
        printf("index file name: %s\n", index_file_name_);
    }

    std::string get_index_file_name() {
        std::string ret(index_file_name_, index_file_name_size_);
        return ret;
    }

    RmRecord insert_value_;     // 索引插入的键
    Rid rid_;
    char* index_file_name_;          // 插入键的索引文件名称
    size_t index_file_name_size_;    // 索引文件名称的大小
};

class IxDeleteLogRecord: public LogRecord {
public:
    IxDeleteLogRecord() {
        log_type_ = LogType::IX_DELETE;
        lsn_ = INVALID_LSN;
        log_tot_len_ = LOG_HEADER_SIZE;
        log_tid_ = INVALID_TXN_ID;
        prev_lsn_ = INVALID_LSN;
        index_file_name_ = nullptr;
    }
    IxDeleteLogRecord(txn_id_t txn_id, const char *key, int key_size, const Rid &rid, std::string table_name) 
        : IxDeleteLogRecord() {
        log_tid_ = txn_id;

        delete_value_.size = key_size;
        delete_value_.data = new char[key_size];
        memcpy(delete_value_.data, key, key_size);

        rid_ = rid;
        log_tot_len_ += sizeof(int);
        log_tot_len_ += sizeof(Rid);
        log_tot_len_ += delete_value_.size;
        index_file_name_size_ = table_name.length();
        index_file_name_ = new char[index_file_name_size_];
        memcpy(index_file_name_, table_name.c_str(), index_file_name_size_);
        log_tot_len_ += sizeof(size_t) + index_file_name_size_;
    }

    // 把insert日志记录序列化到dest中
    void serialize(char* dest) const override {
        LogRecord::serialize(dest);
        int offset = OFFSET_LOG_DATA;
        memcpy(dest + offset, &delete_value_.size, sizeof(int));
        offset += sizeof(int);
        memcpy(dest + offset, delete_value_.data, delete_value_.size);
        offset += delete_value_.size;
        memcpy(dest + offset, &rid_, sizeof(Rid));
        offset += sizeof(Rid);
        memcpy(dest + offset, &index_file_name_size_, sizeof(size_t));
        offset += sizeof(size_t);
        memcpy(dest + offset, index_file_name_, index_file_name_size_);
    }
    // 从src中反序列化出一条Insert日志记录
    void deserialize(const char* src) override {
        LogRecord::deserialize(src);  
        delete_value_.Deserialize(src + OFFSET_LOG_DATA);
        int offset = OFFSET_LOG_DATA + delete_value_.size + sizeof(int);
        rid_ = *reinterpret_cast<const Rid*>(src + offset);
        offset += sizeof(Rid);
        index_file_name_size_ = *reinterpret_cast<const size_t*>(src + offset);
        offset += sizeof(size_t);
        index_file_name_ = new char[index_file_name_size_];
        memcpy(index_file_name_, src + offset, index_file_name_size_);
    }
    void format_print() override {
        printf("delete key\n");
        LogRecord::format_print();
        printf("delete_value_: %s\n", delete_value_.data);
        printf("index file name: %s\n", index_file_name_);
    }

    std::string get_index_file_name() {
        std::string ret(index_file_name_, index_file_name_size_);
        return ret;
    }

    RmRecord delete_value_;     // 索引删除的键
    Rid rid_;
    char* index_file_name_;          // 删除键的索引文件名称
    size_t index_file_name_size_;    // 索引文件名称的大小
};

string toString(LogType type) {
    if(type == LogType::ABORT) {
        return "abort";
    }
    else if(type == LogType::begin) {
        return "begin";
    }
    else if(type == LogType::commit) {
        return "commit";
    }
    else if(type == LogType::DELETE) {
        return "delete";
    }
    else if(type == LogType::INSERT) {
        return "insert";
    }
    else if(type == LogType::IX_DELETE) {
        return "ix_delete";
    }
    else if(type == LogType::IX_INSERT) {
        return "ix_insert";
    }
    else if(type == LogType::UPDATE) {
        return "update";
    }
    else {
        std::cout<<"Failed"<<endl;
        return "error log";
    }
}
#include <errno.h>
int get_file_size(const std::string &file_name) {
    struct stat stat_buf;
    int rc = stat(file_name.c_str(), &stat_buf);
    char error_message[1024];
    strncpy(error_message, strerror(errno), sizeof(error_message) - 1);
    error_message[sizeof(error_message) - 1] = '\0';
    std::string error_(error_message);
    cout<<error_<<endl;
    return rc == 0 ? stat_buf.st_size : -1;
}

int main() {
    int log_fd_ = open("../testdb/db.log", O_RDWR);
    int file_size = get_file_size("../testdb/db.log");

    std::cout<<file_size<<endl;

    std::fstream outfile;
    outfile.open("log_out.txt", std::ios::out | std::ios::app);

    char *log_data = new char[std::max(file_size,1)];

    int offset = 0;
    lseek(log_fd_, offset, SEEK_SET);
    ssize_t bytes_read = read(log_fd_, log_data, file_size);

    LogRecord* new_log = new LogRecord();
    while(file_size !=0 ){
        new_log->deserialize(log_data+offset);

        outfile <<"lsn:"<<to_string(new_log->lsn_)<<" "<<toString(new_log->log_type_)<<" TID:"<<to_string(new_log->log_tid_)<<' ';

        if(new_log->log_type_ == LogType::begin){
            outfile<<'\n';
        }
        else if(new_log->log_type_ == LogType::commit){
            outfile<<'\n';
        }
        else if(new_log->log_type_  == LogType::ABORT){
            outfile<<'\n';
        }
        else if(new_log->log_type_ == LogType::DELETE){
            DeleteLogRecord* del_log_rec = new DeleteLogRecord();
            del_log_rec->deserialize(log_data + offset);
            outfile<<"rid: "<<to_string(del_log_rec->rid_.page_no)<<", "<<to_string(del_log_rec->rid_.slot_no)<<" val: "<<to_string((*(int *)del_log_rec->delete_value_.data))<<" "<<del_log_rec->get_tab_name()<<'\n';
            delete del_log_rec;
        }
        else if(new_log->log_type_ == LogType::INSERT){
            InsertLogRecord* ins_log_rec = new InsertLogRecord();
            ins_log_rec->deserialize(log_data + offset);
            outfile<<"rid: "<<to_string(ins_log_rec->rid_.page_no)<<", "<<to_string(ins_log_rec->rid_.slot_no)<<" val: "<<to_string((*(int *)ins_log_rec->insert_value_.data))<<" "<<ins_log_rec->get_tab_name()<<'\n';
            delete ins_log_rec;
        }
        else if(new_log->log_type_ == LogType::UPDATE){
            UpdateLogRecord* up_log_rec = new UpdateLogRecord();
            up_log_rec->deserialize(log_data + offset);
            outfile<<"rid: "<<to_string(up_log_rec->rid_.page_no)<<", "<<to_string(up_log_rec->rid_.slot_no)<<" old val: "<<to_string((*(int *)up_log_rec->old_value_.data))<<" new val: "<<to_string((*(int *)up_log_rec->update_value_.data))<<" "<<up_log_rec->get_tab_name()<<'\n';
            delete up_log_rec;
        }
        else if(new_log->log_type_ == LogType::IX_INSERT){
            outfile<<'\n';
        }
        else if(new_log->log_type_ == LogType::IX_DELETE){
            outfile<<'\n';
        }
        else {
            outfile<<'\n';
        }

        offset += new_log->log_tot_len_;
        if(offset >= file_size || new_log->log_tot_len_ <= 0){
            break;
        }
    }
    outfile.close();
    delete new_log;

    delete[] log_data;
    return 0;
}